"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@typescript-eslint/utils");
const tsutils = require("tsutils");
const utils = require("../utils");
const rule = utils_1.ESLintUtils.RuleCreator.withoutDocs({
    meta: {
        type: "problem",
        docs: {
            description: "Suggest to remove `async` when not necessary",
            recommended: "warn",
            requiresTypeChecking: true,
        },
        messages: {
            unnecessaryAsync: "This function is not explicitly returning Promise or has no `await` statement, do you want to remove `async` keyword?",
        },
        schema: [],
    },
    defaultOptions: [],
    create(context) {
        const parserServices = utils_1.ESLintUtils.getParserServices(context);
        const checker = parserServices.program.getTypeChecker();
        const sourceCode = context.getSourceCode();
        let scopeInfo = null;
        /**
         * Push the scope info object to the stack.
         */
        function enterFunction(node) {
            scopeInfo = {
                upper: scopeInfo,
                hasAsync: node.async,
                returnsPromise: false,
                hasAwait: false,
            };
        }
        /**
         * Pop the top scope info object from the stack.
         * Also, it reports the function if needed.
         */
        function exitFunction(node) {
            /* istanbul ignore if */ if (!scopeInfo) {
                // this shouldn't ever happen, as we have to exit a function after we enter it
                return;
            }
            if (node.async) {
                if (!scopeInfo.returnsPromise && !scopeInfo.hasAwait) {
                    context.report({
                        node,
                        loc: utils.getFunctionHeadLoc(node, sourceCode),
                        messageId: "unnecessaryAsync",
                    });
                }
            }
            scopeInfo = scopeInfo.upper;
        }
        /**
         * Checks if the node returns a thenable type
         */
        function isThenableType(node) {
            const type = checker.getTypeAtLocation(node);
            return tsutils.isThenableType(checker, node, type);
        }
        /**
         * Marks the current scope as returns Promise explicitly
         */
        function markAsReturnsPromiseExplicitly() {
            if (!scopeInfo) {
                return;
            }
            scopeInfo.returnsPromise = true;
        }
        /**
         * Marks the current scope as has at least 1 await
         */
        function markAsHasAwait() {
            if (!scopeInfo) {
                return;
            }
            scopeInfo.hasAwait = true;
        }
        return {
            FunctionDeclaration: enterFunction,
            FunctionExpression: enterFunction,
            ArrowFunctionExpression: enterFunction,
            "FunctionDeclaration:exit": exitFunction,
            "FunctionExpression:exit": exitFunction,
            "ArrowFunctionExpression:exit": exitFunction,
            // check body-less async arrow function.
            // ignore `async () => await foo` because it's obviously correct
            "ArrowFunctionExpression[async = true] > :not(BlockStatement, AwaitExpression)"(node) {
                const expression = parserServices.esTreeNodeToTSNodeMap.get(node);
                if (expression && isThenableType(expression)) {
                    markAsReturnsPromiseExplicitly();
                }
            },
            ReturnStatement(node) {
                // short circuit early to avoid unnecessary type checks
                if (!scopeInfo || !scopeInfo.hasAsync) {
                    return;
                }
                const { expression } = parserServices.esTreeNodeToTSNodeMap.get(node);
                if (expression && isThenableType(expression)) {
                    markAsReturnsPromiseExplicitly();
                }
            },
            CallExpression(node) {
                var _a;
                // short circuit early to avoid unnecessary type checks
                if (!scopeInfo || !scopeInfo.hasAsync) {
                    return;
                }
                const tsnode = parserServices.esTreeNodeToTSNodeMap.get(node);
                if (tsnode &&
                    isThenableType(tsnode) &&
                    ((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === "AwaitExpression") {
                    const callee = node.callee;
                    if (callee.type === "Identifier") {
                        markAsHasAwait();
                    }
                }
            },
        };
    },
});
exports.default = rule;
